import torch
import torch.nn as nn
import torch.nn.functional as F

class FNN(nn.Module):
    """Fully Connected Neural Network"""

    def __init__(self, input_size, output_size, arch=None, activation='relu', device='cpu',init_zero=False):
        """
        Args:
            input_size (int): Input size for FNN
            output_size (int): Output size for FNN
            arch (list of int): list containing the number of neurons for each hidden layer
            activation (stirng): type of activation function, either 'relu', 
            device (string): computational device
        """
        super().__init__()
        if(arch is None):
            self.neuron_size = [64,32]
        else:
            self.neuron_size = arch
        # Set activation
        self.activation = activation
        self.fc = nn.ModuleList()
        # Assign first layer
        self.fc.append(nn.Linear(input_size, self.neuron_size[0]).to(device))
        if(len(self.neuron_size) == 1):
            # Assign output Layer
            self.fc.append(nn.Linear(self.neuron_size[0], output_size).to(device))
        else:
            for layer in range(1,len(self.neuron_size)):
                self.fc.append(nn.Linear(self.neuron_size[layer-1], self.neuron_size[layer]).to(device))
            # Assign output layer
            self.fc.append(nn.Linear(self.neuron_size[len(self.neuron_size)-1], output_size).to(device))

    def forward(self, x):
        """
        Args:
            x (torch.Tensor): [num_samples, input_size], assumed to be a batch.

        Returns:
            torch.Tensor. Assumed to be a batch.
        """
        for loopA in range(len(self.fc)-1):
            if(self.activation == 'relu'):
                x = F.relu(self.fc[loopA](x))
            elif(self.activation == 'silu'):
                x = F.silu(self.fc[loopA](x))
            elif(self.activation == 'tanh'):
                x = F.tanh(self.fc[loopA](x))
            else:
                print('Invalid activation string.')
                exit(-1)

        # Last layer with linear activation
        x = self.fc[len(self.fc)-1](x)
        return x



#NOTES
#did not change any of the NOFAS parameters to work with multifidelity
#left optimizer and learning rate scheduler the same for both model optimizations, could be adjusted
#did not change annealing parameters for multifidelity since the idea is we will only apply annealing to low fidelity model

import os
import torch
import numpy as np
from linfa.maf import MAF, RealNVP

torch.set_default_tensor_type(torch.DoubleTensor)

class experiment_mf:
    """Defines an instance of variational inference

    This class is the core class of the LINFA library
    and defines all the default hyperparameter values and 
    and functions used for inference. 
    """
    def __init__(self):

        # NF ARCHITECTURE parameters
        self.name              = "Experiment"
        self.input_size        = 3             
        self.flow_type         = 'maf'         
        self.n_blocks_low      = 15     #change n_blocks to n_blocks_low for the low fidelity model
        self.n_blocks_high     = 20     #add n_blocks_high for the high fidelity model (must be greater than n_blocks_low)
        self.hidden_size       = 100           
        self.n_hidden          = 1             
        self.activation_fn     = 'relu'        
        self.input_order       = 'sequential'  
        self.batch_norm_order  = True          

        # NOFAS parameters
        self.run_nofas          = True  
        self.log_interval       = 10    
        self.calibrate_interval = 300   
        self.true_data_num      = 2     
        self.budget             = 216   
        self.surr_pre_it        = 40000 
        self.surr_upd_it        = 6000  
        self.surr_folder        = "./"  
        self.use_new_surr       = True  

        # OPTIMIZER parameters
        self.optimizer       = 'Adam'   
        self.lr_low          = 0.003   #change lr to lr_low for the low fidelity model optimization
        self.lr_high         = 0.003   #add lr_high for the high fidelity model optimization
        self.lr_decay_low    = 0.9999  #change lr_decay to lr_decay_low for the low fidelity model optimization
        self.lr_decay_high   = 0.9999  #add lr_decay_high for the high fidelity model optimization
        self.lr_scheduler    = 'StepLR' 
        self.lr_step_low     = 1000    #change lr_step to lr_step_low for the low fidelity model optimization
        self.lr_step_high    = 1000    #add lr_step_high for the high fidelity model optimization  
        self.batch_size_low  = 500     #change batch_size to batch_size_low for the low fidelity model optimization
        self.batch_size_high = 500     #add batch_size_high for the high fidelity model optimization        
        self.n_iter_low      = 25001   #change n_iter to n_iter_low for the low fidelity model optimization
        self.n_iter_high     = 25001   #add n_iter_high for the high fidelity model optimization

        # ANNEALING parameters
        self.annealing     = True     
        self.scheduler     = 'AdaAnn' 
        # AdaAnn
        self.tol           = 0.001    
        self.t0            = 0.01     
        self.N             = 100      
        self.N_1           = 1000     
        self.T_0           = 500      
        self.T             = 5        
        self.T_1           = 5001     
        self.M             = 1000             
        # Linear scheduler
        self.linear_step  = 0.0001   

        # OUTPUT parameters
        self.output_dir          = './results/' + self.name 
        self.log_file            = 'log.txt'                
        self.seed                = 35435                    
        self.n_sample            = 5000                     
        self.save_interval       = 200                      
        self.store_nf_interval   = 1000                     
        self.store_surr_interval = None                     

        # DEVICE parameters
        self.no_cuda = True #:bool: Flag to use CPU

        # Set device
        self.device = torch.device('cuda:0' if torch.cuda.is_available() and not self.no_cuda else 'cpu')

        # Local pointer to the main components for inference
        self.transform             = None
        self.model_low             = None #change model to model_low for low fidelity model
        self.model_high            = None #add model_high for high fidelity model
        self.model_logdensity_low  = None #change model_logdensity to model_logdensity_low for low fidelity model
        self.model_logdensity_high = None #add model_logdensity_high for high fidelity model
        self.surrogate             = None

        
    def run(self):
        """Runs instance of inference problem        

        """

        # Check is surrogate exists
        if self.run_nofas:
            if not os.path.exists(self.name + ".sur") or not os.path.exists(self.name + ".npz"):
                print("Abort: NoFAS enabled, without surrogate files. \nPlease include the following surrogate files in root directory.\n{}.sur and {}.npz".format(self.name, self.name))
                exit(0)
        # setup file ops
        if not os.path.isdir(self.output_dir):
            os.makedirs(self.output_dir)

        # Save a copy of the data in the result folder so it is handy
        if hasattr(self.model_low,'data'):
          np.savetxt(self.output_dir + '/' + self.name + '_data', self.model.data, newline="\n") #change to save low fidelity data
        if hasattr(self.model_high,'data'):
          np.savetxt(self.output_dir + '/' + self.name + '_data', self.model.data, newline="\n") #add save high fidelity data

        # setup device
        torch.manual_seed(self.seed)
        if self.device.type == 'cuda': torch.cuda.manual_seed(self.seed)

        print('')
        print('--- Running on device: '+ str(self.device))
        print('')

        #CHANGE: set up the first normalizing flow for the low fidelity model optimization here
        # the only change here is the n_blocks_low so the number of layers differ between nf_low and nf_high
        # the remaining parameters need to be the same so the nf_low and nf_high have the same structure
        
        # model
        if self.flow_type == 'maf':
            nf_low = MAF(self.n_blocks_low, self.input_size, self.hidden_size, self.n_hidden, None,
                     self.activation_fn, self.input_order, batch_norm=self.batch_norm_order)
        elif self.flow_type == 'realnvp':  # Under construction
            nf_low = RealNVP(self.n_blocks_low, self.input_size, self.hidden_size, self.n_hidden, None,
                         batch_norm=self.batch_norm_order)
        else:
            raise ValueError('Unrecognized model.')

        
        #CHANGE: set the optimizer and learning rate scheduler for the low fidelity model
        # set nf to nf_low
        # set lr to lr_low
        # set lr_step to lr_step_low
        # set lr_decay to lr_decay_low
        
        nf_low = nf_low.to(self.device)
        if self.optimizer == 'Adam':
            optimizer = torch.optim.Adam(nf_low.parameters(), lr=self.lr_low)
        elif self.optimizer == 'RMSprop':
            optimizer = torch.optim.RMSprop(nf_low.parameters(), lr=self.lr_low)
        else:
            raise ValueError('Unrecognized optimizer.')

        if self.lr_scheduler == 'StepLR':
            scheduler = torch.optim.lr_scheduler.StepLR(optimizer, self.lr_step_low, self.lr_decay_low)
        elif self.lr_scheduler == 'ExponentialLR':
            scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer, self.lr_decay_low)
        else:
            raise ValueError('Unrecognized learning rate scheduler.')

        if self.annealing:
            i = 1
            prev_i = 1
            tvals = []
            t = self.t0
            dt = 0.0
            loglist = []
            while t < 1:
                t = min(1, t + dt)
                tvals = np.concatenate([tvals, np.array([t])])
                self.n_iter_low = self.T
                self.batch_size_low = self.N
                
                if t == self.t0:
                    self.n_iter_low = self.T_0
                if t == 1:
                    self.batch_size_low = self.N_1
                    self.n_iter_low = self.T_1

                while i < prev_i + self.n_iter_low:
                    self.train(nf_low, optimizer, i, loglist, low_fidelity=True, sampling=True, t=t) #added low_fidelity flag
                    if t == 1:
                        scheduler.step()
                    i += 1
                prev_i = i

                if self.scheduler == 'AdaAnn':
                    z0 = nf_low.base_dist.sample([self.M])
                    zk, _ = nf_low(z0)
                    log_qk = self.model_logdensity_low(zk)
                    dt = self.tol / torch.sqrt(log_qk.var())
                    dt = dt.detach()# .numpy()

                if self.scheduler == 'Linear':
                    dt = self.linear_step
        else:
            loglist = []
            for i in range(1, self.n_iter_low+1):                
                self.train(nf_low, optimizer, i, loglist, low_fidelity=True, sampling=True)
                scheduler.step()

                
        #ADD: set up the second normalizing flow for the high fidelity model optimization here
        # the only change here is the n_blocks_high
        
        # model
        if self.flow_type == 'maf':
            nf_high = MAF(self.n_blocks_high, self.input_size, self.hidden_size, self.n_hidden, None,
                     self.activation_fn, self.input_order, batch_norm=self.batch_norm_order)
        elif self.flow_type == 'realnvp':  # Under construction
            nf_high = RealNVP(self.n_blocks_high, self.input_size, self.hidden_size, self.n_hidden, None,
                         batch_norm=self.batch_norm_order)
        else:
            raise ValueError('Unrecognized model.')

        
        #CHANGE: set the optimizer and learning rate scheduler for the high fidelity model
        # reinitializing the optimizer and scheduler, we can rename instead
        # set nf to nf_high
        # set lr to lr_high
        # set lr_step to lr_step_high
        # set lr_decay to lr_decay_high
        
        nf_high = nf_high.to(self.device)
        if self.optimizer == 'Adam':
            optimizer = torch.optim.Adam(nf_high.parameters(), lr=self.lr_high)
        elif self.optimizer == 'RMSprop':
            optimizer = torch.optim.RMSprop(nf_high.parameters(), lr=self.lr_high)
        else:
            raise ValueError('Unrecognized optimizer.')

        if self.lr_scheduler == 'StepLR':
            scheduler = torch.optim.lr_scheduler.StepLR(optimizer, self.lr_step_high, self.lr_decay_high)
        elif self.lr_scheduler == 'ExponentialLR':
            scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer, self.lr_decay_high)
        else:
            raise ValueError('Unrecognized learning rate scheduler.')

        
        #CHANGE: compute the statistics for the batch norm layers
                
        x00 = nf_low.base_dist.sample([10000])
        xkk, _ = nf_low(x00)
        
        mean_vals = xkk.mean(dim=0)
        log_std_vals = torch.log(xkk.std(dim=0)) #log since the batch norm uses log of standard deviation
        
        
        #CHANGE: update the state dictionary for the high fidelity model with the parameters from the low fidelity model
        
        state_dict = nf_high.state_dict()
        state_dict.update(nf_low.state_dict())
        
        
        #CHANGE: update the remaining batch norm layers in the high fidelity model using the computed statistics above
        # the batch norm layers fall on the odd layers of the realnvp framework
        # this is only currently working for realnvp, I still need to look into the maf framework
        
        for j in range(2*self.n_blocks_low + 1, 2*self.n_blocks_high):
        
            if j % 2 == 1:
                state_dict['net.' + str(j) + '.log_gamma'] = log_std_vals
                state_dict['net.' + str(j) + '.beta'] = mean_vals

        
        #CHANGE: update the high fideltiy model with the updated parameters
        nf_high.load_state_dict(state_dict)
        
        
        #CHANGE: determine the length of the parameter list in low fidelity model
        length = len(torch.nn.ParameterList(nf_low.parameters()))
        
        
        #CHANGE: freeze the layers corresponding to the low fidelity model in the new high fidelity model
        # do this by changing the requires_grad parameter to False 
        
        j = 0
        for param in nf_high.parameters():
            param.requires_grad = False
            j+=1
            if j == length:
                break
                
        
        #CHANGE: optimize the high fidelity model
        # did not include annealing here since we are optimizing the high fidelity model, we can always change this
        
        #loglist = []
        for i in range(1, self.n_iter_high+1):                
            self.train(nf_high, optimizer, i, loglist, low_fidelity=False, sampling=True)
            scheduler.step()
                
        print('')
        print('--- Simulation completed!')        

        
    def train(self, nf, optimizer, iteration, log, low_fidelity, sampling=True, t=1): #added a flag to switch between low and high fidelity models
        """Parameter update for normalizing flow and surrogate model

        This is the function where the ELBO loss function is evaluated, 
        the results are saved and the surrogate model is updated.
        
        Args:
            nf (instance of normalizing flow): the normalizing flow architecture used for variational inference
            optimizer (instance of PyTorch optimizer): the selected PyTorch optimizer
            iteration (int): current iteration number
            log (list of lists): stores a log of [iteration, annealing temperature, loss value]
            sampling (bool): Flag indicating the sampling stage
            t (double): current inverse temperature for annealing scheduler

        Returns:
            None
        """

        # Set the normalizing flow in training mode
        nf.train()

        # Evaluate the Jacobian terms in  loss function
        
        if low_fidelity == True:
            x0 = nf.base_dist.sample([self.batch_size_low]) #low fidelity model batch size
            
        else:
            x0 = nf.base_dist.sample([self.batch_size_high]) #high fidelity model batch size
        
        xk, sum_log_abs_det_jacobians = nf(x0)

        # generate and save samples evaluation
        if sampling and iteration % self.save_interval == 0:
            print('--- Saving results at iteration '+str(iteration))
            x00 = nf.base_dist.sample([self.n_sample])
            xkk, _ = nf(x00)
            # Save surrogate grid
            if not(self.surrogate is None):
              np.savetxt(self.output_dir + '/' + self.name + '_grid_' + str(iteration), self.surrogate.grid_record.clone().cpu().numpy(), newline="\n")
            # Save log profile
            np.savetxt(self.output_dir + '/' + self.log_file, np.array(log), newline="\n")
            # Save transformed samples          
            np.savetxt(self.output_dir + '/' + self.name + '_samples_' + str(iteration), xkk.data.clone().cpu().numpy(), newline="\n")
            # Save samples in the original space
            if not(self.transform is None):
              xkk_samples = self.transform.forward(xkk).data.cpu().numpy()
              np.savetxt(self.output_dir + '/' + self.name + '_params_' + str(iteration), xkk_samples, newline="\n")
            else:
              xkk_samples = xkk.data.cpu().numpy()
              np.savetxt(self.output_dir + '/' + self.name + '_params_' + str(iteration), xkk_samples, newline="\n")
            # Save marginal statistics
            np.savetxt(self.output_dir + '/' + self.name + '_marginal_stats_' + str(iteration), np.concatenate((xkk_samples.mean(axis=0).reshape(-1,1),xkk_samples.std(axis=0).reshape(-1,1)),axis=1), newline="\n")
            # Save log density at the same samples
            np.savetxt(self.output_dir + '/' + self.name + '_logdensity_LF_' + str(iteration), self.model_logdensity_low(xkk).data.cpu().numpy(), newline="\n")
            # Save model outputs at the samples - If a model is defined
            if not(self.transform is None):
              stds = torch.abs(self.model.defOut).to(self.device) * self.model.stdRatio
              o00 = torch.randn(x00.size(0), self.model.data.shape[0]).to(self.device)
              noise = o00*stds.repeat(o00.size(0),1)
              if self.surrogate:
                np.savetxt(self.output_dir + '/' + self.name + '_outputs_' + str(iteration), (self.surrogate.forward(xkk) + noise).data.cpu().numpy(), newline="\n")
              else:
                np.savetxt(self.output_dir + '/' + self.name + '_outputs_' + str(iteration), (self.model.solve_t(self.transform.forward(xkk)) + noise).data.cpu().numpy(), newline="\n")

        if torch.any(torch.isnan(xk)):
            print("Error: samples xk are nan at iteration " + str(iteration))
            print(xk)
            np.savetxt(self.output_dir + '/' + self.log_file, np.array(log), newline="\n")
            exit(-1)

        # updating surrogate model
        if self.run_nofas and iteration % self.calibrate_interval == 0 and self.surrogate.grid_record.size(0) < self.budget:
            xk0 = xk[:self.true_data_num, :].data.clone()            
            # print("\n")
            # print(list(self.surrogate.grid_record.size())[0])
            # print(xk0)
            self.surrogate.update(xk0, max_iters=self.surr_upd_it)

        # Free energy bound
        
        
        #CHANGE: need to switch between the low and high fidelity models in the loss function
        # added a flag in the training function depending on which model is being optimized
        # true for low fidelity model, false for high fidelity model
        
        if low_fidelity == True:
            loss = (- torch.sum(sum_log_abs_det_jacobians, 1) - t * self.model_logdensity_low(xk)).mean()
            
        else:
            loss = (- torch.sum(sum_log_abs_det_jacobians, 1) - t * self.model_logdensity_high(xk)).mean()
            
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
        if iteration % self.log_interval == 0:
            print('VI NF (t=%5.3f): it: %7d | loss: %8.3e' % (t,iteration, loss.item()))
            log.append([t, iteration, loss.item()])
        
        # Save state of normalizing flow layers
        if self.store_nf_interval > 0 and iteration % self.store_nf_interval == 0:
            torch.save(nf.state_dict(), self.output_dir + '/' + self.name + "_" + str(iteration) + ".nf")

        if not(self.store_surr_interval is None) and self.store_surr_interval > 0 and iteration % self.store_surr_interval == 0:
            self.surrogate.surrogate_save() # Save surrogate model        




import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.distributions as D
import math
import copy

torch.set_default_tensor_type(torch.DoubleTensor)

# ------------------------
# Model layers and helpers
# ------------------------

def create_masks(input_size, hidden_size, n_hidden, input_order='sequential', input_degrees=None):
    # MADE paper sec 4:
    # degrees of connections between layers -- ensure at most in_degree - 1 connections
    degrees = []

    # set input degrees to what is provided in args (the flipped order of the previous layer in a stack of mades);
    # else init input degrees based on strategy in input_order (sequential or random)
    if input_order == 'sequential':
        degrees += [torch.arange(input_size)] if input_degrees is None else [input_degrees]
        for _ in range(n_hidden + 1):
            degrees += [torch.arange(hidden_size) % (input_size - 1)]
        degrees += [torch.arange(input_size) % input_size - 1] if input_degrees is None else [
            input_degrees % input_size - 1]

    elif input_order == 'random':
        degrees += [torch.randperm(input_size)] if input_degrees is None else [input_degrees]
        for _ in range(n_hidden + 1):
            min_prev_degree = min(degrees[-1].min().item(), input_size - 1)
            degrees += [torch.randint(min_prev_degree, input_size, (hidden_size,))]
        min_prev_degree = min(degrees[-1].min().item(), input_size - 1)
        degrees += [torch.randint(min_prev_degree, input_size, (input_size,)) - 1] if input_degrees is None else [
            input_degrees - 1]

    # construct masks
    masks = []
    for (d0, d1) in zip(degrees[:-1], degrees[1:]):
        masks += [(d1.unsqueeze(-1) >= d0.unsqueeze(0))]

    return masks, degrees[0]


class MaskedLinear(nn.Linear):
    """ MADE building block layer """

    def __init__(self, input_size, n_outputs, mask, cond_label_size=None):
        super().__init__(input_size, n_outputs)
        # self.weight.data.fill_(1.00)
        # self.bias.data.fill_(0.00)
        # nn.init.kaiming_uniform_(self.weight, mode='fan_out', nonlinearity='relu')
        # nn.init.kaiming_normal_(self.weight, mode='fan_out', nonlinearity='relu')

        self.register_buffer('mask', mask)

        self.cond_label_size = cond_label_size
        if cond_label_size is not None:
            self.cond_weight = nn.Parameter(torch.rand(n_outputs, cond_label_size) / math.sqrt(cond_label_size))

    def forward(self, x, y=None):
        out = F.linear(x, self.weight * self.mask, self.bias)
        if y is not None:
            out = out + F.linear(y, self.cond_weight)
        return out

    def extra_repr(self):
        return 'in_features={}, out_features={}, bias={}'.format(
            self.in_features, self.out_features, self.bias is not None
        ) + (self.cond_label_size != None) * ', cond_features={}'.format(self.cond_label_size)


class LinearMaskedCoupling(nn.Module):
    """ Modified RealNVP Coupling Layers per the MAF paper """

    def __init__(self, input_size, hidden_size, n_hidden, mask, cond_label_size=None):
        super().__init__()

        self.register_buffer('mask', mask)

        # scale function
        s_net = [nn.Linear(input_size + (cond_label_size if cond_label_size is not None else 0), hidden_size)]
        for _ in range(n_hidden):
            s_net += [nn.Tanh(), nn.Linear(hidden_size, hidden_size)]

        s_net += [nn.Tanh(), nn.Linear(hidden_size, input_size)]  # Default nn.Tanh()
        self.s_net = nn.Sequential(*s_net)

        # translation function
        self.t_net = copy.deepcopy(self.s_net)
        # replace Tanh with ReLU's per MAF paper
        for i in range(len(self.t_net)):
            if not isinstance(self.t_net[i], nn.Linear): self.t_net[i] = nn.ReLU()

    def forward(self, x, y=None):
        # apply mask
        mx = x * self.mask

        # run through model
        s = self.s_net(mx if y is None else torch.cat([y, mx], dim=1))
        t = self.t_net(mx if y is None else torch.cat([y, mx], dim=1))
        u = mx + (1 - self.mask) * (x - t) * torch.exp(
            -s)  # cf RealNVP eq 8 where u corresponds to x (here we're modeling u)

        log_abs_det_jacobian = - (
                1 - self.mask) * s  # log det du/dx; cf RealNVP 8 and 6; note, sum over input_size done at model log_prob

        return u, log_abs_det_jacobian

    def inverse(self, u, y=None):
        # apply mask
        mu = u * self.mask

        # run through model
        s = self.s_net(mu if y is None else torch.cat([y, mu], dim=1))
        t = self.t_net(mu if y is None else torch.cat([y, mu], dim=1))
        x = mu + (1 - self.mask) * (u * s.exp() + t)  # cf RealNVP eq 7

        log_abs_det_jacobian = (1 - self.mask) * s  # log det dx/du

        return x, log_abs_det_jacobian


class BatchNorm(nn.Module):
    """ RealNVP BatchNorm layer """

    def __init__(self, input_size, momentum=0.9, eps=1e-5):
        super().__init__()
        self.momentum = momentum
        self.eps = eps

        self.log_gamma = nn.Parameter(torch.zeros(input_size))
        self.beta = nn.Parameter(torch.zeros(input_size))

        self.register_buffer('running_mean', torch.zeros(input_size))
        self.register_buffer('running_var', torch.ones(input_size))

    def forward(self, x, cond_y=None):
        if self.training:
            self.batch_mean = x.mean(0)
            self.batch_var = x.var(0)  # note MAF paper uses biased variance estimate; ie x.var(0, unbiased=False)

            # update running mean
            self.running_mean.mul_(self.momentum).add_(self.batch_mean.data * (1 - self.momentum))
            self.running_var.mul_(self.momentum).add_(self.batch_var.data * (1 - self.momentum))

            mean = self.batch_mean
            var = self.batch_var
        else:
            mean = self.running_mean
            var = self.running_var

        # compute normalized input (cf original batch norm paper algo 1)
        x_hat = (x - mean) / torch.sqrt(var + self.eps)
        y = self.log_gamma.exp() * x_hat + self.beta

        # compute log_abs_det_jacobian (cf RealNVP paper)
        log_abs_det_jacobian = self.log_gamma - 0.5 * torch.log(var + self.eps)
        # print('in sum log var {:6.3f} ; out sum log var {:6.3f}; sum log det {:8.3f}; mean log_gamma {:5.3f}; mean
        # beta {:5.3f}'.format( (var + self.eps).log().sum().data.numpy(), y.var(0).log().sum().data.numpy(),
        # log_abs_det_jacobian.mean(0).item(), self.log_gamma.mean(), self.beta.mean()))
        return y, log_abs_det_jacobian.expand_as(x)

    def inverse(self, y, cond_y=None):
        if self.training:
            mean = self.batch_mean
            var = self.batch_var
        else:
            mean = self.running_mean
            var = self.running_var

        x_hat = (y - self.beta) * torch.exp(-self.log_gamma)
        x = x_hat * torch.sqrt(var + self.eps) + mean

        log_abs_det_jacobian = 0.5 * torch.log(var + self.eps) - self.log_gamma

        return x, log_abs_det_jacobian.expand_as(x)


class FlowSequential(nn.Sequential):
    """ Container for layers of a normalizing flow """

    def forward(self, x, y):
        sum_log_abs_det_jacobians = 0
        for module in self:
            x, log_abs_det_jacobian = module(x, y)
            sum_log_abs_det_jacobians = sum_log_abs_det_jacobians + log_abs_det_jacobian
        return x, sum_log_abs_det_jacobians

    def inverse(self, u, y):
        sum_log_abs_det_jacobians = 0
        for module in reversed(self):
            u, log_abs_det_jacobian = module.inverse(u, y)
            sum_log_abs_det_jacobians = sum_log_abs_det_jacobians + log_abs_det_jacobian
        return u, sum_log_abs_det_jacobians


# ------
# Models
# ------

class MADE(nn.Module):
    def __init__(self, input_size, hidden_size, n_hidden, cond_label_size=None, activation='relu',
                 input_order='sequential', input_degrees=None):
        """
        Args:
            input_size -- scalar; dim of inputs
            hidden_size -- scalar; dim of hidden layers
            n_hidden -- scalar; number of hidden layers
            activation -- str; activation function to use
            input_order -- str or tensor; variable order for creating the autoregressive masks (sequential|random)
                            or the order flipped from the previous layer in a stack of mades
            conditional -- bool; whether model is conditional
        """
        super().__init__()
        # base distribution for calculation of log prob under the model
        self.register_buffer('base_dist_mean', torch.zeros(input_size))
        self.register_buffer('base_dist_var', torch.ones(input_size))

        # create masks
        masks, self.input_degrees = create_masks(input_size, hidden_size, n_hidden, input_order, input_degrees)

        # setup activation
        if activation == 'relu':
            activation_fn = nn.ReLU()
        elif activation == 'tanh':
            activation_fn = nn.Tanh()
        elif activation == 'sigmoid':
            activation_fn = nn.Sigmoid()
        else:
            raise ValueError('Check activation function.')

        # construct model
        self.net_input = MaskedLinear(input_size, hidden_size, masks[0], cond_label_size)
        self.net = []
        for m in masks[1:-1]:
            self.net += [activation_fn, MaskedLinear(hidden_size, hidden_size, m)]
        self.net += [activation_fn, MaskedLinear(hidden_size, 2 * input_size, masks[-1].repeat(2, 1))]
        self.net = nn.Sequential(*self.net)

    @property
    def base_dist(self):
        return D.Normal(self.base_dist_mean, self.base_dist_var)

    def forward(self, x, y=None):
        # MAF eq 4 -- return mean and log std
        m, loga = self.net(self.net_input(x, y)).chunk(chunks=2, dim=1)
        u = (x - m) * torch.exp(-loga)
        # MAF eq 5
        log_abs_det_jacobian = - loga
        return u, log_abs_det_jacobian

    def inverse(self, u, y=None, sum_log_abs_det_jacobians=None):
        # MAF eq 3
        D = u.shape[1]
        x = torch.zeros_like(u)
        # run through reverse model
        for i in self.input_degrees:
            m, loga = self.net(self.net_input(x, y)).chunk(chunks=2, dim=1)
            x[:, i] = u[:, i] * torch.exp(loga[:, i]) + m[:, i]
        log_abs_det_jacobian = loga
        return x, log_abs_det_jacobian

    def log_prob(self, x, y=None):
        u, log_abs_det_jacobian = self.forward(x, y)
        return torch.sum(self.base_dist.log_prob(u) + log_abs_det_jacobian, dim=1)


class MAF(nn.Module):
    def __init__(self, n_blocks, input_size, hidden_size, n_hidden, cond_label_size=None, activation='relu',
                 input_order='sequential', batch_norm=True):
        super().__init__()
        # base distribution for calculation of log prob under the model
        self.register_buffer('base_dist_mean', torch.zeros(input_size))
        self.register_buffer('base_dist_var', torch.ones(input_size))

        # construct model
        modules = []
        self.input_degrees = None
        for i in range(n_blocks):
            modules += [
                MADE(input_size, hidden_size, n_hidden, cond_label_size, activation, input_order, self.input_degrees)]
            self.input_degrees = modules[-1].input_degrees.flip(0)
            modules += batch_norm * [BatchNorm(input_size)]

        self.net = FlowSequential(*modules)

    @property
    def base_dist(self):
        return D.Normal(self.base_dist_mean, self.base_dist_var)

    def forward(self, x, y=None):
        return self.net(x, y)

    def inverse(self, u, y=None):
        return self.net.inverse(u, y)

    def log_prob(self, x, y=None):
        u, sum_log_abs_det_jacobians = self.forward(x, y)
        return torch.sum(self.base_dist.log_prob(u) + sum_log_abs_det_jacobians, dim=1)


class RealNVP(nn.Module):
    def __init__(self, n_blocks, input_size, hidden_size, n_hidden, cond_label_size=None, batch_norm=True):
        super().__init__()

        # base distribution for calculation of log prob under the model
        self.register_buffer('base_dist_mean', torch.zeros(input_size))
        self.register_buffer('base_dist_var', torch.ones(input_size))

        # construct model
        modules = []
        mask = torch.arange(input_size) % 2
        for i in range(n_blocks):
            modules += [LinearMaskedCoupling(input_size, hidden_size, n_hidden, mask, cond_label_size)]
            mask = 1 - mask
            modules += batch_norm * [BatchNorm(input_size)]

        self.net = FlowSequential(*modules)

    @property
    def base_dist(self):
        return D.Normal(self.base_dist_mean, self.base_dist_var)

    def forward(self, x, y=None):
        return self.net(x, y)

    def inverse(self, u, y=None):
        return self.net.inverse(u, y)

    def log_prob(self, x, y=None):
        u, sum_log_abs_det_jacobians = self.forward(x, y)
        return torch.sum(self.base_dist.log_prob(u) + sum_log_abs_det_jacobians, dim=1)


import numpy as np
import torch

torch.set_default_tensor_type(torch.DoubleTensor)

class Trivial:
    def __init__(self,device='cpu'):
        # Init parameters
        self.defParam = torch.Tensor([[3.0, 5.0]]).to(device)
        self.RM = torch.Tensor([[1.0, 1.0],
                                [1.0, -1.0]]).to(device)
        self.stdRatio = 0.05
        self.defOut = self.solve_t(self.defParam)
        self.data = None

    def genDataFile(self, dataSize=50, dataFileName="../../resource/data/data_trivial.txt", store=True):
        def_out = self.solve_t(self.defParam)[0]
        print(def_out)
        self.data = def_out + self.stdRatio * torch.abs(def_out) * torch.normal(0, 1, size=(dataSize, 2))
        self.data = self.data.t().detach().numpy()
        if store: np.savetxt(dataFileName, self.data)
        return self.data

    def solve_t(self, params):
        z1, z2 = torch.chunk(params, chunks=2, dim=1)
        x = torch.cat((z1 ** 3 / 10, torch.exp(z2 / 3)), 1)
        return torch.matmul(x, self.RM)

# GEN DATA
if __name__ == '__main__':
  
  model = Trivial()
  model.genDataFile(50)


import torch
import itertools
import numpy as np

class PhysChem(object):

    def __init__(self, var_inputs):
        
        # variable inputs
        self.var_in = torch.tensor(list(itertools.product(*var_inputs)))
        
        # calibration inputs
        self.defParams = torch.tensor([[1e3, -21e3]]) # standard presssure (MPa) and energy of ads. (J/mol)
        self.limits = [[5e2, 1.5e3],[-30e3, -15e3]] # range on defParams
                
        ## model constants
        self.RConst = torch.tensor(8.314) # universal gas constant (J/ mol/ K)
        self.data = None # dataset of model
        self.stdRatio = 0.05 # standard deviation ratio
        self.defOut = self.solve_t(self.defParams)
        
    def solve_t(self, cal_inputs):

        num_batch = len(cal_inputs)
        num_vars = len(self.var_in)
        
        # unpack variable inputs
        T, P = torch.chunk(self.var_in, chunks = 2, dim = 1)
        T = T.repeat(1, num_batch)
        P = P.repeat(1, num_batch)

        # unpack calibration inputs
        p0Const, eConst = torch.chunk(cal_inputs, chunks = 2, dim = 1) # split cal_inputs into two chunks along the second dimension
        p0Const = p0Const.repeat(1, num_vars).t()
        eConst  = eConst.repeat(1, num_vars).t()
        
        # compute equilibrium constant
        kConst = 1/p0Const* torch.exp(-eConst / self.RConst / T)
        
        # compute surface coverage fraction
        cov_frac = kConst*P/(1 + kConst*P)

        # Return coverages        
        return cov_frac

    def solve_true(self, cal_inputs):

        num_batch = len(cal_inputs)
        num_vars = len(self.var_in)

        # unpack variable input
        T, P = torch.chunk(self.var_in, chunks = 2, dim = 1)
        T = T.repeat(1, num_batch)
        P = P.repeat(1, num_batch)

        # unpack ground truth calibration input (adsorption site 1)
        p01Const, e1Const = torch.chunk(cal_inputs, chunks = 2, dim = 1)
        p01Const = p01Const.repeat(1, num_vars).t()
        e1Const  = e1Const.repeat(1, num_vars).t()

        # specify adsorption site two parameters
        p02Const, e2Const, lambda1Const, lambda2Const = torch.tensor(5000.0), torch.tensor(-22000.0), torch.tensor(1.0), torch.tensor(0.5)
        p02Const = p02Const.repeat(1, num_vars).t()
        e2Const  = e2Const.repeat(1, num_vars).t()

        # compute equilibrium constant of site one
        k1Const = 1.0/p01Const * torch.exp(-e1Const / self.RConst / T)

        # compute equilibrium constant of site two
        k2Const = 1.0/p02Const * torch.exp(-e2Const / self.RConst / T)

        # compute surface coverage fraction for two adsorption sites with different equilibrium conditions
        cov_frac = lambda1Const * (k1Const*P/(1 + k1Const*P)) + lambda2Const * (k2Const*P/(1 + k2Const*P))

        # Return
        return cov_frac

    
    def genDataFile(self, dataFileNamePrefix='observations', use_true_model=True, store=True, num_observations=10):

        # solve model
        if(use_true_model):
            def_out = self.solve_true(self.defParams)
        else:
            def_out = self.solve_t(self.defParams)
        
        # get standard deviation
        stdDev = self.stdRatio * torch.abs(def_out)
        
        # add noise to coverage data
        coverage = def_out.repeat(1,num_observations)

        for loopA in range(num_observations):
            noise = torch.randn((len(coverage),1))*stdDev
            coverage[:,loopA] = coverage[:,loopA] + noise.flatten()
        
        if store:

            # specify file name
            file_name = f'{dataFileNamePrefix}.csv'

            with open(file_name, 'w') as file:

                # write file
                file.write('temperature, pressure, coverage \n')

                # specify row names
                for loopA in range(self.var_in.size(0)):
                    for loopB in range(self.var_in.size(1)):
                        file.write(f'{self.var_in[loopA,loopB]},')                        
                    for loopB in range(coverage.size(1)-1):
                        file.write(f'{coverage[loopA,loopB].item()},')
                    file.write(f'{coverage[loopA,coverage.size(1)-1].item()}\n')

# TEST SURROGATE
if __name__ == '__main__':

    # Set variable grid
    var_grid = [[350.0, 400.0, 450.0],
                [1.0, 2.0, 3.0, 4.0, 5.0]]

    # Create model
    model = PhysChem(var_grid)
    
    # Generate data
    model.genDataFile(num_observations=50)


    


from linfa.run_experiment import experiment
from linfa.transform import Transformation
import torch
import random
import numpy as np

# Import rcr model
from linfa.models.discrepancy_models import PhysChem

def run_test():

    exp = experiment()
    exp.name = "lf_no_disc_lf_data_tp1"
    exp.flow_type           = 'maf'         # str: Type of flow (default 'realnvp')
    exp.n_blocks            = 15            # int: Number of hidden layers   
    exp.hidden_size         = 100           # int: Hidden layer size for MADE in each layer (default 100)
    exp.n_hidden            = 1             # int: Number of hidden layers in each MADE
    exp.activation_fn       = 'relu'        # str: Activation function used (default 'relu')
    exp.input_order         = 'sequential'  # str: Input oder for create_mask (default 'sequential')
    exp.batch_norm_order    = True          # bool: Order to decide if batch_norm is used (default True)
    exp.save_interval       = 1000          # int: How often to sample from normalizing flow
    
    exp.input_size          = 2             # int: Dimensionalty of input (default 2)
    exp.batch_size          = 200           # int: Number of samples generated (default 100)
    exp.true_data_num       = 2             # double: Number of true model evaluted (default 2)
    exp.n_iter              = 25001         # int: Number of iterations (default 25001)
    exp.lr                  = 0.001         # float: Learning rate (default 0.003)
    exp.lr_decay            = 0.9999        # float:  Learning rate decay (default 0.9999)
    exp.log_interal         = 10            # int: How often to show loss stat (default 10)

    exp.run_nofas           = False         # normalizing flow with adaptive surrogate
    exp.surrogate_type      = 'discrepancy' # type of surrogate we are using
    exp.surr_pre_it         = 1000          # int: Number of pre-training iterations for surrogate model
    exp.surr_upd_it         = 1000          # int: Number of iterations for the surrogate model update

    exp.annealing           = False
    exp.calibrate_interval  = 300           # int: How often to update the surrogate model (default 1000)
    exp.budget              = 216           # int: Total number of true model evaulations
    exp.surr_folder         = "./" 
    exp.use_new_surr        = True

    exp.output_dir = './results/' + exp.name
    exp.log_file = 'log.txt'
    
    exp.seed = random.randint(0, 10 ** 9)   # int: Random seed used
    exp.n_sample = 5000                     # int: Batch size to generate final results/plots
    exp.no_cuda = True                      # Running on CPU by default but teste on CUDA

    exp.optimizer = 'RMSprop'
    exp.lr_scheduler = 'ExponentialLR'

    exp.device = torch.device('cuda:0' if torch.cuda.is_available() and not exp.no_cuda else 'cpu')

    # Define transformation
    trsf_info = [['linear', -4.0, 4.0, 500.0, 1500.0],
                 ['linear', -4.0, 4.0, -30000.0, -15000.0]]
    trsf = Transformation(trsf_info)
    
    # Apply the transformation
    exp.transform = trsf

    # Add temperatures and pressures for each evaluation
    variable_inputs = [[350.0],
                       [1.0]]

    # Define model
    langmuir_model = PhysChem(variable_inputs)
    # Assign as experiment model
    exp.model = langmuir_model

    # Read data
    exp.model.data = np.loadtxt('observations.csv', delimiter = ',', skiprows = 1)

    if(len(exp.model.data.shape) < 2):
        exp.model.data = np.expand_dims(exp.model.data, axis=0)
    
    # No surrogate
    exp.surrogate = None

    # Define log density
    def log_density(calib_inputs, model, surrogate, transform):
        
        # Compute transformation by log Jacobian
        adjust = transform.compute_log_jacob_func(calib_inputs)

        # Initialize negative log likelihood
        total_nll = torch.zeros((calib_inputs.size(0), 1))
        
        # HAD TO MOVE THIS UP BEFORE EVALUATE DISCREPANCY            
        # Evaluate model response - (num_var x num_batch)
        modelOut = langmuir_model.solve_t(transform.forward(calib_inputs)).t()

        # Evaluate discrepancy
        if (surrogate is None):
            discrepancy = torch.zeros(modelOut.shape).t()
        else:
            # (num_var)
            discrepancy = surrogate.forward(model.var_in)
        
        # Get the absolute values of the standard deviation (num_var)
        stds = langmuir_model.defOut * langmuir_model.stdRatio
        
        # Get data - (num_var x num_obs)
        Data = torch.tensor(langmuir_model.data[:,2:]).to(exp.device)
        num_obs = Data.size(1)
        
        # Evaluate log-likelihood:
        # Loop on the available observations
        for loopA in range(num_obs):
            l1 = -0.5 * np.prod(langmuir_model.data.shape) * np.log(2.0 * np.pi)
            
            # TODO: generalize to multiple inputs
            l2 = (-0.5 * langmuir_model.data.shape[1] * torch.log(torch.prod(stds))).item()
            l3 = -0.5 * torch.sum(((modelOut + discrepancy.t() - Data[:,loopA].unsqueeze(0)) / stds.t())**2, dim = 1)
            
            # Compute negative ll (num_batch x 1)
            negLL = -(l1 + l2 + l3) # sum contributions
            res = -negLL.reshape(calib_inputs.size(0), 1) # reshape
        
            # Accumulate
            total_nll += res
                
        # Return log-likelihood
        return total_nll/num_obs + adjust

    # Assign log density model
    exp.model_logdensity = lambda x: log_density(x, exp.model, exp.surrogate, exp.transform)

    # Run VI
    exp.run()

def generate_data():

    # Set variable grid
    var_grid = [[350.0],
                [1.0]]

    # Create model
    model = PhysChem(var_grid)
    
    # Generate data
    model.genDataFile(use_true_model=False,num_observations=1)

# Main code
if __name__ == "__main__":
    
    generate_data()
    
    run_test()










import numpy as np
import matplotlib.pyplot as plt

folder = 'results/test_out_with_disc/'
step_num = 501
plot_all_LF = False


obs_file = 'test_out_with_disc_data'
lf_file = 'test_out_with_disc_outputs_lf_' + str(step_num)
lf_discr_file = 'test_out_with_disc_outputs_lf+discr_' + str(step_num)
discr_file = 'test_out_with_disc_outputs_discr_' + str(step_num)

obs = np.loadtxt(folder + obs_file)
lf = np.loadtxt(folder + lf_file)
lf_discr = np.loadtxt(folder + lf_discr_file)
discr = np.loadtxt(folder + discr_file)

# Separate T,P from coverage
tp = obs[:,:2]
obs = obs[:,2:]

fig = plt.figure()
ax = fig.add_subplot(projection='3d')
# True coverage
ax.scatter(tp[:,0],tp[:,1],obs,label='obs',color='b')

if(plot_all_LF):
	# LF
	for loopA in range(len(tp)):
		if(loopA == 0):
			ax.scatter(tp[:,0],tp[:,1],lf[:,loopA],color='r',label='LF',alpha=0.6)
		else:
			ax.scatter(tp[:,0],tp[:,1],lf[:,loopA],color='r',alpha=0.6)
	# DISCR
	for loopA in range(len(tp)):
		if(loopA == 0):
			ax.scatter(tp[:,0],tp[:,1],lf_discr[:,loopA],color='m',label='LF+DISC',alpha=0.6)
		else:
			ax.scatter(tp[:,0],tp[:,1],lf_discr[:,loopA],color='m',alpha=0.6)
else:
	# Plot average LF and Discrepancy
	ax.scatter(tp[:,0],tp[:,1],np.mean(lf,axis=1),color='r',label='LF',alpha=0.6)		
	ax.scatter(tp[:,0],tp[:,1],np.mean(lf_discr,axis=1),marker='D',color='m',label='LF+DISC',alpha=0.6)

plt.legend()

plt.show()




import numpy as np
import matplotlib.pyplot as plt
import itertools

fs=14

T = 350.0
P = 1.0

p0 = 1000.0
e0 = -21000.0

variable_inputs = [[350.0, 375.0 ,400.0, 425.0, 450.0],
                   [1.0, 2.0, 3.0, 4.0, 5.0]]

tp_pairs = list(itertools.product(*variable_inputs))

e_const = np.linspace(-30000,-15000,1000)

plt.figure(figsize=(5,3))
for t,p in tp_pairs:
	k_0 = (1/p0)*np.exp(-e0/(8.314*t))
	p_const = (1/k_0)*np.exp(-e_const/(8.314*t))
	plt.plot(p_const,e_const,label='t: '+str(t)+', p:'+str(p))
plt.scatter(p0,e0)
#plt.legend()
plt.xlim([500,1500])
plt.ylim([-23000,-18000])
plt.xlabel(r'$p_{0}$',fontsize=fs)
plt.ylabel(r'$e$')
plt.tight_layout()
plt.show()




import math
import torch
import numpy as np
from functools import partial

# Identity
def id_fun(x):
    return x
def id_jac(x):
    return torch.zeros_like(x)

# Tanh
def tanh_fun(x,m1,t1,m2,t2):
    return torch.tanh((x - m1) / t1 * 3.0) * t2 + m2
def tanh_jac(x,m1,t1,m2,t2):
    return torch.log(1.0 - torch.tanh((x - m1) / t1 * 3.0) ** 2) + np.log(t2) + np.log(3.0) - np.log(t1)

# Linear
def lin_fun(x,a,b,c,d):
    return  (x - a) / (b - a) * (d - c) + c
def lin_jac(x,a,b,c,d):
    return torch.tensor([np.log(d - c) - np.log(b - a)]).repeat(x.size(0), 1)

# Exponential
def exp_fun(x,a,b,c,d):
    return torch.exp((x - a) / (b - a) * (np.log(d) - np.log(c)) + np.log(c))
def exp_jac(x,a,b,c,d):
    return (x - a) / (b - a) * (np.log(d) - np.log(c)) + np.log(c) + np.log(np.log(d) - np.log(c)) - np.log(b - a)

class Transformation(torch.nn.Module):

    def __init__(self, func_info=None):
        super().__init__()
        self.funcs     = []
        self.log_jacob = []
        self.n = len(func_info)
        for func, a, b, c, d in func_info:
            if func == "identity":
                self.funcs.append(id_fun)
                self.log_jacob.append(id_jac)
            elif func == "tanh":
                m1 = (a + b) / 2
                t1 = (b - a) / 2
                m2 = (c + d) / 2
                t2 = (d - c) / 2
                self.funcs.append(partial(tanh_fun,m1=m1,t1=t1,m2=m2,t2=t2))
                self.log_jacob.append(partial(tanh_jac,m1=m1,t1=t1,m2=m2,t2=t2))
            elif func == "linear":
                self.funcs.append(partial(lin_fun,a=a,b=b,c=c,d=d))
                self.log_jacob.append(partial(lin_jac,a=a,b=b,c=c,d=d))
            elif func == "exp":
                self.funcs.append(partial(exp_fun,a=a,b=b,c=c,d=d))
                self.log_jacob.append(partial(exp_jac,a=a,b=b,c=c,d=d))
    
    def forward(self, z):
        """
        
        from desired scale to original scale, evaluate original samples.

        Args:
            z (torch.Tensor): samples in desired scale.

        Returns:
            torch.Tensor: samples in original scale.
        """
        if z.size(1) != self.n:
            raise ValueError("Inconsistent size. Got {}, should be {}".format(z.size(1), self.n))
        zi = torch.chunk(z, chunks=self.n, dim=1)
        x = []
        for zz, func in zip(zi, self.funcs):
            x.append(func(zz))

        return torch.cat(x, dim=1)

    def compute_log_jacob_func(self, z):
        """
        
        from desired scale to original scale, compute log absolute determinant of Jacobian matrix.

        Args:
            z (torch.Tensor): samples in desired scale.

        Returns:
            torch.Tensor: samples in original scale.
        """
        if z.size(1) != self.n:
            raise ValueError("Inconsistent size. Got {}, should be {}".format(z.size(1), self.n))
        zi = torch.chunk(z, chunks=self.n, dim=1)
        jacob_res = 0
        for zz, log_jacob_func in zip(zi, self.log_jacob):
            jacob_res += log_jacob_func(zz)

        return jacob_res
    
def test_transform(tets_num):
  
  import matplotlib.pyplot as plt
  
  # Define the transformation params
  if(tets_num == 0):
    a =-7.0
    b = 7.0
    c = 100.0
    d = 1500.0
    trsf_info = [['tanh',a,b,c,d]]
  elif(tets_num == 1):
    a =-7.0
    b = 7.0
    c = math.exp(-8.0)
    d = math.exp(-5.0)
    trsf_info = [['exp',a,b,c,d]]

  # Create transformation
  trsf = Transformation(trsf_info)

  # Create uniform parameterization in [a,b]
  z_vals = torch.from_numpy(np.linspace(a,b,1000)).reshape((-1,1))

  # Eval transformation
  x_vals = trsf.forward(z_vals)
  
  # Eval Jacobian
  x_grad = trsf.compute_log_jacob_func(z_vals)

  # Plot the transformation
  plt.subplot(1,2,1)
  plt.title('Forward map')
  plt.plot(z_vals.numpy(),x_vals.numpy(),'b-')  
  plt.xlim([a,b])
  plt.ylim([c,d])
  plt.subplot(1,2,2)
  plt.title('Jacobian')
  plt.plot(z_vals.numpy(),x_grad.numpy(),'r-')
  plt.xlim([a,b])
  plt.show()

def test_gradient():
  # Set transformation parameters
  # a =-7.0
  a = 0.0
  b = 7.0
  c = 100.0
  d = 1500.0

  trsf_info = [['tanh',a,b,c,d]]
  # trsf_info = [['identity',0,0,0,0]]
  # trsf_info = [['exp',a,b,c,d]]
  # trsf_info = [['linear',a,b,c,d]]

  trsf = Transformation(trsf_info)

  # Create uniform parameterization in [a,b]
  z_vals = torch.from_numpy(np.linspace(a,b,10)).reshape((-1,1))
  z_vals.requires_grad=True
  
  print('z_vals ',z_vals)

  # Eval transformation
  x_vals = trsf.forward(z_vals)

  print('x_vals ',x_vals)
  
  # Eval Jacobian
  x_grad = trsf.compute_log_jacob_func(z_vals)

  x_vals[0].backward()

  print('z_vals.grad ',torch.log(z_vals.grad))
  print('x_grad ',x_grad)

# TEST TRANSFORMATION
if __name__ == '__main__':

  # test_transform(0)
  test_gradient()


from linfa.run_experiment import experiment
from linfa.transform import Transformation
from linfa.discrepancy import Discrepancy
import torch
import random
import numpy as np

# Import rcr model
from linfa.models.discrepancy_models import PhysChem

def run_test():

    exp = experiment()
    exp.name = "only_discr_ex"
    exp.flow_type           = 'maf'         # str: Type of flow (default 'realnvp')
    exp.n_blocks            = 15            # int: Number of hidden layers   
    exp.hidden_size         = 100           # int: Hidden layer size for MADE in each layer (default 100)
    exp.n_hidden            = 1             # int: Number of hidden layers in each MADE
    exp.activation_fn       = 'relu'        # str: Activation function used (default 'relu')
    exp.input_order         = 'sequential'  # str: Input oder for create_mask (default 'sequential')
    exp.batch_norm_order    = True          # bool: Order to decide if batch_norm is used (default True)
    exp.save_interval       = 1000          # int: How often to sample from normalizing flow
    
    exp.input_size          = 2             # int: Dimensionalty of input (default 2)
    exp.batch_size          = 200           # int: Number of samples generated (default 100)
    exp.true_data_num       = 2             # double: Number of true model evaluted (default 2)
    exp.n_iter              = 2501          # int: Number of iterations (default 25001)
    exp.lr                  = 0.001         # float: Learning rate (default 0.003)
    exp.lr_decay            = 0.9999        # float:  Learning rate decay (default 0.9999)
    exp.log_interal         = 10            # int: How often to show loss stat (default 10)

    exp.run_nofas           = True          # normalizing flow with adaptive surrogate
    exp.surrogate_type      = 'discrepancy' # type of surrogate we are using
    exp.surr_pre_it         = 10000         # int: Number of pre-training iterations for surrogate model
    exp.surr_upd_it         = 1000          # int: Number of iterations for the surrogate model update

    exp.annealing           = False
    exp.calibrate_interval  = 300           # int: How often to update the surrogate model (default 1000)
    exp.budget              = 216           # int: Total number of true model evaulations
    exp.surr_folder         = "./" 
    exp.use_new_surr        = True

    exp.output_dir = './results/' + exp.name
    exp.log_file = 'log.txt'
    
    exp.seed = random.randint(0, 10 ** 9)   # int: Random seed used
    exp.n_sample = 5000                     # int: Batch size to generate final results/plots
    exp.no_cuda = True                      # Running on CPU by default but teste on CUDA

    exp.optimizer = 'RMSprop'
    exp.lr_scheduler = 'ExponentialLR'

    exp.device = torch.device('cuda:0' if torch.cuda.is_available() and not exp.no_cuda else 'cpu')

    # Add temperatures and pressures for each evaluation
    variable_inputs = [[350.0, 400.0, 450.0],
                       [1.0, 2.0, 3.0, 4.0, 5.0]]

    # Define model
    langmuir_model = PhysChem(variable_inputs)
    # Assign as experiment model
    exp.model = langmuir_model
    # Read data
    exp.model.data = np.loadtxt('observations.csv', delimiter = ',', skiprows = 1)

    if(len(exp.model.data.shape) < 2):
        exp.model.data = np.expand_dims(exp.model.data, axis=0)
    
    # Form tensors for variables and results in observations
    var_grid_in = torch.tensor(exp.model.data[:,:2])
    var_grid_out = torch.tensor(exp.model.data[:,2:])

    # Define surrogate
    if(exp.run_nofas):

        # Create new discrepancy
        exp.surrogate = Discrepancy(model_name=exp.name, 
                                    lf_model=exp.model.solve_t,
                                    input_size=exp.model.var_in.size(1),
                                    output_size=1,
                                    var_grid_in=var_grid_in,
                                    var_grid_out=var_grid_out)
        # Initially tune on the default values of the calibration variables
        exp.surrogate.update(langmuir_model.defParams, exp.surr_pre_it, 0.03, 0.9999, 100, store=True)
        # Load the surrogate
        exp.surrogate.surrogate_load()
    else:
        exp.surrogate = None


    # Verify discrepancy training
    discr = exp.surrogate.forward(var_grid_in)
    lf_model = exp.model.solve_t(exp.model.defParams)
    print('%15s %15s' % ('Model+Discr','Observations'))
    for loopA in range(var_grid_in.size(0)):
        print('%15.6f %15.6f' % (lf_model[loopA][0].item()+discr[loopA][0].item(),var_grid_out[loopA][0].item()))

def generate_data(use_true_model=False, num_observations=50):

    # Set variable grid
    var_grid = [[350.0, 400.0, 450.0],
                [1.0, 2.0, 3.0, 4.0, 5.0]]

    # Create model
    model = PhysChem(var_grid)
    
    # Generate data
    model.genDataFile(use_true_model=use_true_model,num_observations=num_observations)

# Main code
if __name__ == "__main__":
    
    generate_data(use_true_model = True, num_observations = 1)
    run_test()







